#ifndef INC_CONJUNTOS
#define INC_CONJUNTOS
#include <cstdlib>
/**
 * La clase Conjunto define objetos que representan conjuntos de
 * elementos
 */
template <class T>
   class Conjunto {
   public:
   // Constructoras
      Conjunto();          // Representa el conjunto Vaco
      Conjunto(const T &); // representa el conjunto Unitario
      ~Conjunto();         // Destructor
      Conjunto(const Conjunto<T> &); // Constructor de copia
      Conjunto<T> & operator =(const Conjunto<T> &); // Asignacin de conjuntos
      Conjunto<T> operator +(const Conjunto<T> &) const; // Operacin Unin 
      Conjunto<T> operator *(const Conjunto<T> &) const; // Oper. Interseccin
      Conjunto<T> operator -(const Conjunto<T> &) const; // Oper. Diferencia
   
   // Observadoras
      bool operator ==(const  Conjunto<T> &) const; // Igualdad de conjuntos
      bool incluido(const Conjunto<T> &) const;     // Operacin Inclusin 
      bool pertenece(const T &) const;              // Operacin Pertenece
      int cardinal() const;                         // Operacin Cardinal
   private:
      /**
       * El tipo Nodo es privado, ya que su razn de ser es servir de
       * "ladrillo" para formar la estructura del conjunto como
       * lista simplemente encadenada
       */
      struct Nodo {
         T elemento;
         Nodo *siguiente;
         Nodo(T e, Nodo *s=NULL):elemento(e),siguiente(s){} // Constructor
      };
      typedef Nodo *PNodo;
   
      PNodo elementos;  // Puntero al primer nodo de la lista encadenada
      int numElementos; // Mantiene el cardinal del conjunto
   
      /**
       * Realiza una copia de una lista simplemente encadenada
       */
      static PNodo copiaElementos(PNodo p);
   
      /**
       * Libera la memoria ocupada por los elementos de una lista encadenada
       */
      void eliminaElementos();
   
   };

template<class T>
   void Conjunto<T>::eliminaElementos() {
      PNodo pAux;
   
      while (elementos) {
         pAux = elementos;
         elementos = elementos->siguiente;
         delete pAux;
      }
   };

template<class T>
   typename Conjunto<T>::PNodo Conjunto<T>::copiaElementos(Conjunto<T>::PNodo p) {         
      PNodo actual_p, primCopia, ultCopia, nuevo;
   
      actual_p = p;
      primCopia = NULL;
      while (actual_p) {
         nuevo = new Nodo(actual_p->elemento,actual_p->siguiente);
         if (primCopia == NULL) 
            primCopia = nuevo;
         else 
            ultCopia->siguiente = nuevo;
         ultCopia = nuevo;
         actual_p = actual_p->siguiente;
      }
      return primCopia;
   };


/**
 * Constructor del conjunto Vaco (constructor por defecto)
 */
template <class T>
   Conjunto<T>::Conjunto() {
      elementos = NULL;
      numElementos = 0;
   };

/**
 * Constructor del conjunto Unitario
 */
template <class T>
   Conjunto<T>::Conjunto(const T &e) {
      elementos = new Nodo(e);
      numElementos = 1;
   };

/**
 * Destructor: destruye la lista que representa al conjunto
 */
template <class T>
   Conjunto<T>::~Conjunto() {
      eliminaElementos();
   };

/**
 * Constructor de copia: para pasar un conjunto como parmetro por  
 * valor o devolverlo como valor de retorno
 */
template <class T>
   Conjunto<T>::Conjunto(const Conjunto<T> &c) {
      elementos = copiaElementos(c.elementos);
      numElementos = c.numElementos;
   };

/**
 * Asignacin de conjuntos con semntica de copia
 */
template <class T>
   Conjunto<T> & Conjunto<T>::operator =(const Conjunto<T> &c) {
      if (this != &c) // Si no es una autoasignacin
      {
         eliminaElementos();
         elementos = copiaElementos(c.elementos);
         numElementos = c.numElementos;
      }
      return *this;
   };

/**
 * La unin de dos conjuntos se realiza copiando el primero en un
 * conjunto auxiliar en el que luego se insertan, mediante un
 * recorrido del segundo todos los elementos de ste que no
 * pertenecen al primero.
 */
template <class T>
   Conjunto<T> Conjunto<T>::operator +(const Conjunto<T> &c) const {
      Conjunto<T> cAux = c; // Copia del primer conjunto
      PNodo pTemp, pAux = elementos;
   
      while (pAux) {
         if (!c.pertenece(pAux->elemento)) {
            pTemp = cAux.elementos;
            cAux.elementos = new Nodo(pAux->elemento,pTemp);
            cAux.numElementos = cAux.numElementos + 1;
         }
         pAux = pAux->siguiente;
      }
      return cAux;
   };

template <class T>
   Conjunto<T> Conjunto<T>::operator *(const Conjunto<T> &c) const{
      Conjunto<T> cAux;
      PNodo pTemp, pAux = elementos;
   
      while (pAux) {
         if (c.pertenece(pAux->elemento)) {
            pTemp = cAux.elementos;
            cAux.elementos = new Nodo(pAux->elemento,pTemp);
            cAux.numElementos = cAux.numElementos + 1;
         }
         pAux = pAux->siguiente;
      }
      return cAux;
   };

template <class T>
   Conjunto<T> Conjunto<T>::operator -(const Conjunto<T> &c) const{
      Conjunto<T> cAux;
      PNodo pTemp, pAux = elementos;
   
      while (pAux) {
         if (!c.pertenece(pAux->elemento)) {
            pTemp = cAux.elementos;
            cAux.elementos = new Nodo(pAux->elemento,pTemp);
            cAux.numElementos = cAux.numElementos + 1;
         }
         pAux = pAux->siguiente;
      }
      return cAux;
   };

template <class T>
   bool Conjunto<T>::operator ==(const Conjunto<T> &c) const{
      return ((cardinal() == c.cardinal()) && 
             ((this == &c) || incluido(c))
             );
   };

template <class T>
   bool Conjunto<T>::incluido(const Conjunto<T> &c) const{
      PNodo pAux = elementos;
   
      if (numElementos > c.numElementos) {
         return false;
      }
      else {
         while (pAux) {
            if (!c.pertenece(pAux->elemento)) {
               return false;
            }
            pAux = pAux->siguiente;
         }
         return true;
      }
   };

template <class T>
   bool Conjunto<T>::pertenece(const T &e) const{
      PNodo pAux = elementos;
   
      while (pAux) {
         if (pAux->elemento == e) {
            return true;
         }
         pAux = pAux->siguiente;
      }
      return false;
   };

template <class T>
   int Conjunto<T>::cardinal() const{
      return numElementos;
   };

#endif
